
-- Can we create a type that holds values like (String "and" (Integer "or" Double"))
data Test = Test String (Integer | Double) -- "or" can't be used this way

-- Can we have a return type that returns values that are (Integer "or" Double)
f :: Integer -> Integer -> (Integer | Double)  -- "or" can't be used this way
g :: Integer -> Integer -> (Integer , Double)  -- "and" is OK
g x y = (x `div` y, fromIntegral x / fromIntegral y)

-- If you look at the syntax diagram for defining data types,
-- you can see that it does not allow using "or" in the ways
-- that we are trying to use it above.

-- To solve these problems, Haskell has the Either data type.
data Test = Test String (Either Integer Double) deriving Show
test1 = Test "hello" (Left 3)
test2 = Test "cat"   (Right 2.3)

-- This is like writing
--    x * (y + z)
-- as
--    x * add(y, z)
-- where we use an add function instead of the + operator.

-- When we want a "sum type" to have higher precedence than
-- a "product type", the "sum type" needs to be put in an
-- Either constructor.

-- Here is an example of a "or" return type.
f :: Integer -> Integer -> Either Integer Double
f x y | (x `mod` y == 0) = Left (x `div` y)
f x y | otherwise = Right (fromIntegral x / fromIntegral y)

x1 = f 4 3  -- returns a Right Double
x2 = f 4 2  -- returns a Left Integer
x3 = f 4 0  -- causes an error


-- Here is an example of a "or" return type in a Maybe.
h :: Integer -> Integer -> Maybe (Either Integer Double)
h _ 0 = Nothing
h x y | (x `mod` y == 0) = Just (Left (x `div` y))
h x y | otherwise = Just (Right (fromIntegral x / fromIntegral y))

y1 = h 4 0  -- returns a Nothing
y2 = h 4 2  -- returns a Just Left Integer
y3 = h 4 3  -- returns a Just Right Double
